/* $Id: timecoded.c,v 1.2 1998/12/28 15:55:11 mas Rel $
 * Copyright 1998 Marc Andr Selig.
 * All rights reserved.
 * Timecode reader daemon.
 * Voraussetzungen: Timecode-Lesekarte Alpermann+Velte AV PCL-1/2;
 *	Root-Permissions (fr Hardware-Port-Zugriff).
 * telnet to me at INET_PORT and tell me anything; I will return a timecode
 *	read from the board, or "t/c missing" if there is no board
 * logging is done via syslog
 */

/*
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* Bugs:

 * none currently known

 */

/* Ideas for future releases:

 * The program could automatically detect if the board is missing and
 * switch to "no-board operation" instead of failing.
 
 */



#include <stdio.h>
#include <errno.h>
#include <unistd.h>
#include <getopt.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <stdlib.h>
#include <syslog.h>
#include "socklib.h"

#if defined __GLIBC__ && __GLIBC__ >= 2
#include <sys/io.h>
#endif

#define SYSLOG_FACILITY LOG_LOCAL6
	/* facility to use for syslog */

#define TC_PORT_START 0x240	/* TC reader card address */
#define TC_PORT_REQUEST (TC_PORT_START+14)	/* request port */
#define TC_PORT_ADDRESS (TC_PORT_START+16)	/* address port */
#define TC_PORT_SIGNAL 128	/* signal to emit on request port */
	/* 0x80 = 128 means "read asynchronously"; 0x40 = 64 would be
	 * "read synchronously", i.e. wait for the next valid time code */

static uid_t euid, ruid;

int tc_card_present;



void do_setuid (void)
/* Restore the effective UID to its original value. */
{
  int status;
     
#ifdef _POSIX_SAVED_IDS
  status = setuid (euid);
#else
  status = setreuid (ruid, euid);
#endif
  if (status < 0) {
    syslog (SYSLOG_FACILITY | LOG_ERR, "couldn't set uid.");
    syslog (SYSLOG_FACILITY | LOG_ERR, "exiting now.");
    exit (status);
  }
}



void undo_setuid (void)
/* Set the effective UID to the real UID. */
{
  int status;
     
#ifdef _POSIX_SAVED_IDS
  status = setuid (ruid);
#else
  status = setreuid (euid, ruid);
#endif
  if (status < 0) {
    syslog (SYSLOG_FACILITY | LOG_ERR, "couldn't set uid.");
    syslog (SYSLOG_FACILITY | LOG_ERR, "exiting now.");
    exit (status);
  }
}



void watch_dog(int x) {
  syslog (SYSLOG_FACILITY | LOG_ERR, "cannot read time code from port 0x%x!", TC_PORT_START);
  syslog (SYSLOG_FACILITY | LOG_ERR, "exiting now.");
  exit(1);
}



void get_timecode (char *buffer) {	/* buffer MUST be 14 bytes minimum */
  /* reads the current timecode from the reader board */
  /* this assumes we have access to the relevant ports */
  outb(TC_PORT_SIGNAL, TC_PORT_REQUEST); /* Tell the board to scan */
  if (tc_card_present) {
    alarm(1);
    do {}	/* Wait until scan result has been written into registers. */
    while (inb(TC_PORT_REQUEST));    /* Request port will then revert to 0. */
    alarm(0);
    snprintf (buffer, 14, "%02x:%02x:%02x:%02x\n",
	      inb(TC_PORT_START+3),
	      inb(TC_PORT_START+2),
	      inb(TC_PORT_START+1),
	      inb(TC_PORT_START));
    /* userbits would be at TC_PORT_START+8 .. TC_PORT_START+11 */
  } else
    strcpy(buffer, "t/c missing\n");
}



int main (int argc, char *argv[]) {
  int sock;
  int fd_iterate, n;
  fd_set active_fd_set, read_fd_set;
  struct sockaddr_in clientname;
  size_t size;
  char timecode[16];

  /* Save the real and effective user IDs.  */
  ruid = getuid ();
  euid = geteuid ();
  /* Give ourselves permission for hardware port access. */
  /* NB: This requires root privileges! */
  ioperm (TC_PORT_START, 32, 255);
  /* Change our persona to that of the user. */
  undo_setuid ();

  tc_card_present = 1;

  while ((n = getopt(argc, argv, "nh?")) != -1)
    switch (n) {
    case 'n':	/* no time code reader card installed */
      tc_card_present = 0;
      break;

    case 'h':
    case '?':
      printf("\n$Id: timecoded.c,v 1.2 1998/12/28 15:55:11 mas Rel $\n\n");

      printf("Copyright 1998 Marc Andr Selig.\n");
      printf("timecoded comes with ABSOLUTELY NO WARRANTY; for details see file COPYING.\n");
      printf("This is free software, and you are welcome to redistribute it under\n");
      printf("certain conditions; see file COPYING for details.\n\n");

      printf("Usage: timecoded [-n] [-h|-?]\n");
      printf("  -n\n");
      printf("\tdisable time code reading\n");
      printf("\t(allows program to accept events on computers without time code reader\n");
      printf("\tcard, but makes data gathering useless on computers with such a card)\n");
      printf("  -h|-?\n");
      printf("\tprint this help screen\n\n");
      exit(0);
    }

  /* open syslog */
  openlog("timecoded", 0, LOG_LOCAL6);
  syslog (SYSLOG_FACILITY | LOG_INFO, "startup.");

  /* Watch dog: Program exits if no time code can be read, for example
   * when there is no reader card installed. */
  signal(SIGALRM, watch_dog);

  /* Initialize timecode reader card */
  outb(0, TC_PORT_ADDRESS);
  /* This tells the card which of its registers we want to access.  In
   * this case, we choose register 0, which is the so-called "mixed
   * register" for reader 1. */
  
  /* Create the socket and set it up to accept connections. */
  sock = make_socket (INET_PORT);
  if (listen (sock, 1) < 0) {
      syslog (SYSLOG_FACILITY | LOG_ERR, "listen: %m");
      syslog (SYSLOG_FACILITY | LOG_ERR, "exiting now.");
      exit (EXIT_FAILURE);
  }

  /* Initialize the set of active sockets. */
  FD_ZERO (&active_fd_set);
  FD_SET (sock, &active_fd_set);

  /* main program loop */
  while (1) {
    /* Block until input arrives on one or more active sockets. */
    read_fd_set = active_fd_set;
    if (select (FD_SETSIZE, &read_fd_set, NULL, NULL, NULL) < 0) {
      syslog (SYSLOG_FACILITY | LOG_ERR, "select: %m");
      syslog (SYSLOG_FACILITY | LOG_ERR, "exiting now.");
      exit (EXIT_FAILURE);
    }

    /* Service all the sockets with input pending. */
    for (fd_iterate = 0; fd_iterate < FD_SETSIZE; ++fd_iterate)
      if (FD_ISSET (fd_iterate, &read_fd_set)) {
	if (fd_iterate == sock) {
	  /* Connection request on original socket. */
	  int new;
	  size = sizeof (clientname);
	  new = accept (sock,
			(struct sockaddr *) &clientname,
			&size);
	  if (new < 0) {
	    syslog (SYSLOG_FACILITY | LOG_ERR, "accept: %m");
	    syslog (SYSLOG_FACILITY | LOG_ERR, "exiting now.");
	    exit (EXIT_FAILURE);
	  }
	  syslog (SYSLOG_FACILITY | LOG_INFO,
		   "connect from host %s, port %hd.",
		   inet_ntoa (clientname.sin_addr),
		   ntohs (clientname.sin_port));
	  FD_SET (new, &active_fd_set);
	} else {
	  /* Data arriving on an already-connected socket. */
	  char buf [512];
	  if (read_from_socket (fd_iterate, buf, sizeof(buf)) < 0) {
	    close (fd_iterate);
	    FD_CLR (fd_iterate, &active_fd_set);
	  } else {
	    get_timecode(timecode);
	    write_to_socket(fd_iterate, timecode);
	  }
	}
      }
  }
  exit(0);
}
